@@ -1,6 +1,7 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*-
  *
  * Copyright (C) 2009-2010 Red Hat, Inc.
+ * Copyright (c) 2013 Canonical Limited
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -31,9 +32,6 @@
 #include <unistd.h>
 #include <errno.h>
 #include <sys/types.h>
-#ifdef HAVE_UTMPX_H
-#include <utmpx.h>
-#endif
 
 #include <glib.h>
 #include <glib/gi18n.h>
@@ -42,44 +40,16 @@
 #include <gio/gio.h>
 #include <polkit/polkit.h>
 
+#include "user-classify.h"
+#include "wtmp-helper.h"
 #include "daemon.h"
 #include "util.h"
 
 #define PATH_PASSWD "/etc/passwd"
 #define PATH_SHADOW "/etc/shadow"
-#define PATH_NOLOGIN "/sbin/nologin"
-#define PATH_FALSE "/bin/false"
+#define PATH_GROUP "/etc/group"
 #define PATH_GDM_CUSTOM "/etc/gdm/custom.conf"
 
-static const char *default_excludes[] = {
-        "bin",
-        "root",
-        "daemon",
-        "adm",
-        "lp",
-        "sync",
-        "shutdown",
-        "halt",
-        "mail",
-        "news",
-        "uucp",
-        "operator",
-        "nobody",
-        "nobody4",
-        "noaccess",
-        "postgres",
-        "pvm",
-        "rpm",
-        "nfsnobody",
-        "pcap",
-        "mysql",
-        "ftp",
-        "games",
-        "man",
-        "at",
-        NULL
-};
-
 enum {
         PROP_0,
         PROP_DAEMON_VERSION
@@ -90,18 +60,20 @@ struct DaemonPrivate {
         GDBusProxy *bus_proxy;
 
         GHashTable *users;
-        GHashTable *exclusions;
 
         User *autologin;
 
         GFileMonitor *passwd_monitor;
         GFileMonitor *shadow_monitor;
+        GFileMonitor *group_monitor;
         GFileMonitor *gdm_monitor;
+        GFileMonitor *wtmp_monitor;
 
         guint reload_id;
         guint autologin_id;
 
         PolkitAuthority *authority;
+        GHashTable *extension_ifaces;
 };
 
 typedef struct passwd * (* EntryGeneratorFunc) (GHashTable *, gpointer *);
@@ -157,215 +129,10 @@ error_get_type (void)
   return etype;
 }
 
-gboolean
-daemon_local_user_is_excluded (Daemon *daemon, const gchar *username, const gchar *shell)
-{
-        int ret;
-
-        if (g_hash_table_lookup (daemon->priv->exclusions, username)) {
-                return TRUE;
-        }
-
-        ret = FALSE;
-
-        if (shell != NULL) {
-                char *basename, *nologin_basename, *false_basename;
-
-#ifdef HAVE_GETUSERSHELL
-                char *valid_shell;
-
-                ret = TRUE;
-                setusershell ();
-                while ((valid_shell = getusershell ()) != NULL) {
-                        if (g_strcmp0 (shell, valid_shell) != 0)
-                                continue;
-                        ret = FALSE;
-                }
-                endusershell ();
+#ifndef HAVE_FGETPWENT
+#include "fgetpwent.c"
 #endif
 
-                basename = g_path_get_basename (shell);
-                nologin_basename = g_path_get_basename (PATH_NOLOGIN);
-                false_basename = g_path_get_basename (PATH_FALSE);
-
-                if (shell[0] == '\0') {
-                        ret = TRUE;
-                } else if (g_strcmp0 (basename, nologin_basename) == 0) {
-                        ret = TRUE;
-                } else if (g_strcmp0 (basename, false_basename) == 0) {
-                        ret = TRUE;
-                }
-
-                g_free (basename);
-                g_free (nologin_basename);
-                g_free (false_basename);
-        }
-
-        return ret;
-}
-
-#ifdef HAVE_UTMPX_H
-
-typedef struct {
-        int frequency;
-        gint64 time;
-        GList *previous_logins;
-} UserAccounting;
-
-typedef struct {
-        gchar  *id;
-        gint64  login_time;
-        gint64  logout_time;
-} UserPreviousLogin;
-
-typedef struct {
-        GHashTable *login_hash;
-        GHashTable *logout_hash;
-} WTmpGeneratorState;
-
-static struct passwd *
-entry_generator_wtmp (GHashTable *users,
-                      gpointer   *state)
-{
-        GHashTable *login_hash, *logout_hash;
-        struct utmpx *wtmp_entry;
-        GHashTableIter iter;
-        gpointer key, value;
-        struct passwd *pwent;
-        User *user;
-        WTmpGeneratorState *state_data;
-        GVariantBuilder *builder, *builder2;
-        GList *l;
-
-        if (*state == NULL) {
-                /* First iteration */
-#ifdef UTXDB_LOG
-                if (setutxdb (UTXDB_LOG, NULL) != 0) {
-                        return NULL;
-                }
-#else
-                utmpxname (_PATH_WTMPX);
-                setutxent ();
-#endif
-                *state = g_new (WTmpGeneratorState, 1);
-                state_data = *state;
-                state_data->login_hash = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
-                state_data->logout_hash = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
-        }
-
-        /* Every iteration */
-        state_data = *state;
-        login_hash = state_data->login_hash;
-        logout_hash = state_data->logout_hash;
-        while ((wtmp_entry = getutxent ())) {
-                UserAccounting    *accounting;
-                UserPreviousLogin *previous_login;
-
-                if (wtmp_entry->ut_type == BOOT_TIME) {
-                        /* Set boot time for missing logout records */
-                        g_hash_table_iter_init (&iter, logout_hash);
-                        while (g_hash_table_iter_next (&iter, &key, &value)) {
-                                previous_login = (UserPreviousLogin *) value;
-
-                                if (previous_login->logout_time == 0) {
-                                        previous_login->logout_time = wtmp_entry->ut_tv.tv_sec;
-                                }
-                        }
-                        g_hash_table_remove_all (logout_hash);
-                } else if (wtmp_entry->ut_type == DEAD_PROCESS) {
-                        /* Save corresponding logout time */
-                        if (g_hash_table_lookup_extended (logout_hash, wtmp_entry->ut_line, &key, &value)) {
-                                previous_login = (UserPreviousLogin *) value;
-                                previous_login->logout_time = wtmp_entry->ut_tv.tv_sec;
-
-                                g_hash_table_remove (logout_hash, previous_login->id);
-                        }
-                }
-
-                if (wtmp_entry->ut_type != USER_PROCESS) {
-                        continue;
-                }
-
-                if (wtmp_entry->ut_user[0] == 0) {
-                        continue;
-                }
-
-                pwent = getpwnam (wtmp_entry->ut_user);
-                if (pwent == NULL) {
-                        continue;
-                }
-
-                if (!g_hash_table_lookup_extended (login_hash,
-                                                   wtmp_entry->ut_user,
-                                                   &key, &value)) {
-                        accounting = g_new (UserAccounting, 1);
-                        accounting->frequency = 0;
-                        accounting->previous_logins = NULL;
-
-                        g_hash_table_insert (login_hash, g_strdup (wtmp_entry->ut_user), accounting);
-                } else {
-                        accounting = value;
-                }
-
-                accounting->frequency++;
-                accounting->time = wtmp_entry->ut_tv.tv_sec;
-
-                /* Add zero logout time to change it later on logout record */
-                previous_login = g_new (UserPreviousLogin, 1);
-                previous_login->id = g_strdup (wtmp_entry->ut_line);
-                previous_login->login_time = wtmp_entry->ut_tv.tv_sec;
-                previous_login->logout_time = 0;
-                accounting->previous_logins = g_list_prepend (accounting->previous_logins, previous_login);
-
-                g_hash_table_insert (logout_hash, g_strdup (wtmp_entry->ut_line), previous_login);
-
-                return pwent;
-        }
-
-        /* Last iteration */
-        endutxent ();
-
-        g_hash_table_iter_init (&iter, login_hash);
-        while (g_hash_table_iter_next (&iter, &key, &value)) {
-                UserAccounting    *accounting = (UserAccounting *) value;
-                UserPreviousLogin *previous_login;
-
-                user = g_hash_table_lookup (users, key);
-                if (user == NULL) {
-                        for (l = accounting->previous_logins; l != NULL; l = l->next) {
-                                previous_login = l->data;
-                                g_free (previous_login->id);
-                        }
-                        g_list_free (accounting->previous_logins);
-                        continue;
-                }
-
-                g_object_set (user, "login-frequency", accounting->frequency, NULL);
-                g_object_set (user, "login-time", accounting->time, NULL);
-
-                builder = g_variant_builder_new (G_VARIANT_TYPE ("a(xxa{sv})"));
-                for (l = g_list_last (accounting->previous_logins); l != NULL; l = l->prev) {
-                        previous_login = l->data;
-
-                        builder2 = g_variant_builder_new (G_VARIANT_TYPE ("a{sv}"));
-                        g_variant_builder_add (builder2, "{sv}", "type", g_variant_new_string (previous_login->id));
-                        g_variant_builder_add (builder, "(xxa{sv})", previous_login->login_time, previous_login->logout_time, builder2);
-                        g_variant_builder_unref (builder2);
-                        g_free (previous_login->id);
-                }
-                g_object_set (user, "login-history", g_variant_new ("a(xxa{sv})", builder), NULL);
-                g_variant_builder_unref (builder);
-                g_list_free (accounting->previous_logins);
-        }
-
-        g_hash_table_unref (login_hash);
-        g_hash_table_unref (logout_hash);
-        g_free (state_data);
-        *state = NULL;
-        return NULL;
-}
-#endif /* HAVE_UTMPX_H */
-
 static struct passwd *
 entry_generator_fgetpwent (GHashTable *users,
                            gpointer   *state)
@@ -411,7 +178,7 @@ entry_generator_cachedir (GHashTable *users,
 
         /* First iteration */
         if (*state == NULL) {
-                *state = dir = g_dir_open (USERDIR, 0, &error);
+                *state = g_dir_open (USERDIR, 0, &error);
                 if (error != NULL) {
                         if (!g_error_matches (error, G_FILE_ERROR, G_FILE_ERROR_NOENT))
                                 g_warning ("couldn't list user cache directory: %s", USERDIR);
@@ -440,7 +207,7 @@ entry_generator_cachedir (GHashTable *users,
                 if (regular) {
                         pwent = getpwnam (name);
                         if (pwent == NULL)
-                                g_debug ("user '%s' in cache dir but not presen on system", name);
+                                g_debug ("user '%s' in cache dir but not present on system", name);
                         else
                                 return pwent;
                 }
@@ -456,7 +223,7 @@ entry_generator_cachedir (GHashTable *users,
                 key_file = g_key_file_new ();
                 if (g_key_file_load_from_file (key_file, filename, 0, NULL))
                         user_update_from_keyfile (user, key_file);
-                g_key_file_free (key_file);
+                g_key_file_unref (key_file);
                 g_free (filename);
         }
 
@@ -481,7 +248,7 @@ load_entries (Daemon             *daemon,
                         break;
 
                 /* Skip system users... */
-                if (daemon_local_user_is_excluded (daemon, pwent->pw_name, pwent->pw_shell)) {
+                if (!user_classify_is_human (pwent->pw_uid, pwent->pw_name, pwent->pw_shell, NULL)) {
                         g_debug ("skipping user: %s", pwent->pw_name);
                         continue;
                 }
@@ -546,9 +313,7 @@ reload_users (Daemon *daemon)
                 g_hash_table_add (local, name);
 
         /* Now add/update users from other sources, possibly non-local */
-#ifdef HAVE_UTMPX_H
-        load_entries (daemon, users, entry_generator_wtmp);
-#endif
+        load_entries (daemon, users, wtmp_helper_entry_generator);
         load_entries (daemon, users, entry_generator_cachedir);
 
         /* Mark which users are local, which are not */
@@ -629,7 +394,7 @@ reload_autologin_timeout (Daemon *daemon)
         }
 
         if (enabled) {
-                g_debug ("automatic login is enabled for '%s'\n", name);
+                g_debug ("automatic login is enabled for '%s'", name);
                 if (daemon->priv->autologin != user) {
                         g_object_set (user, "automatic-login", TRUE, NULL);
                         daemon->priv->autologin = g_object_ref (user);
@@ -637,7 +402,7 @@ reload_autologin_timeout (Daemon *daemon)
                 }
         }
         else {
-                g_debug ("automatic login is disabled\n");
+                g_debug ("automatic login is disabled");
         }
 
         g_free (name);
@@ -679,11 +444,11 @@ queue_reload_autologin (Daemon *daemon)
 }
 
 static void
-on_passwd_monitor_changed (GFileMonitor      *monitor,
-                           GFile             *file,
-                           GFile             *other_file,
-                           GFileMonitorEvent  event_type,
-                           Daemon            *daemon)
+on_users_monitor_changed (GFileMonitor      *monitor,
+                          GFile             *file,
+                          GFile             *other_file,
+                          GFileMonitorEvent  event_type,
+                          Daemon            *daemon)
 {
         if (event_type != G_FILE_MONITOR_EVENT_CHANGED &&
             event_type != G_FILE_MONITOR_EVENT_CREATED) {
@@ -708,74 +473,70 @@ on_gdm_monitor_changed (GFileMonitor      *monitor,
         queue_reload_autologin (daemon);
 }
 
+typedef void FileChangeCallback (GFileMonitor      *monitor,
+                                 GFile             *file,
+                                 GFile             *other_file,
+                                 GFileMonitorEvent  event_type,
+                                 Daemon            *daemon);
+
+static GFileMonitor *
+setup_monitor (Daemon             *daemon,
+               const gchar        *path,
+               FileChangeCallback *callback)
+{
+        GError *error = NULL;
+        GFile *file;
+        GFileMonitor *monitor;
+
+        if (!path) {
+                return NULL;
+        }
+
+        file = g_file_new_for_path (path);
+        monitor = g_file_monitor_file (file,
+                                       G_FILE_MONITOR_NONE,
+                                       NULL,
+                                       &error);
+        if (monitor != NULL) {
+                g_signal_connect (monitor,
+                                  "changed",
+                                  G_CALLBACK (callback),
+                                  daemon);
+        } else {
+                g_warning ("Unable to monitor %s: %s", path, error->message);
+                g_error_free (error);
+        }
+        g_object_unref (file);
+
+        return monitor;
+}
+
 static void
 daemon_init (Daemon *daemon)
 {
-        gint i;
-        GFile *file;
-        GError *error;
-
         daemon->priv = DAEMON_GET_PRIVATE (daemon);
 
-        daemon->priv->exclusions = g_hash_table_new_full (g_str_hash,
-                                                          g_str_equal,
-                                                          g_free,
-                                                          NULL);
-
-        for (i = 0; default_excludes[i] != NULL; i++) {
-                g_hash_table_insert (daemon->priv->exclusions,
-                                     g_strdup (default_excludes[i]),
-                                     GUINT_TO_POINTER (TRUE));
-        }
+        daemon->priv->extension_ifaces = daemon_read_extension_ifaces ();
 
         daemon->priv->users = create_users_hash_table ();
 
-        file = g_file_new_for_path (PATH_PASSWD);
-        daemon->priv->passwd_monitor = g_file_monitor_file (file,
-                                                            G_FILE_MONITOR_NONE,
-                                                            NULL,
-                                                            &error);
-        g_object_unref (file);
-        file = g_file_new_for_path (PATH_SHADOW);
-        daemon->priv->shadow_monitor = g_file_monitor_file (file,
-                                                            G_FILE_MONITOR_NONE,
-                                                            NULL,
-                                                            &error);
-        g_object_unref (file);
-        file = g_file_new_for_path (PATH_GDM_CUSTOM);
-        daemon->priv->gdm_monitor = g_file_monitor_file (file,
-                                                         G_FILE_MONITOR_NONE,
-                                                         NULL,
-                                                         &error);
-        g_object_unref (file);
+        daemon->priv->passwd_monitor = setup_monitor (daemon,
+                                                      PATH_PASSWD,
+                                                      on_users_monitor_changed);
+        daemon->priv->shadow_monitor = setup_monitor (daemon,
+                                                      PATH_SHADOW,
+                                                      on_users_monitor_changed);
+        daemon->priv->group_monitor = setup_monitor (daemon,
+                                                     PATH_GROUP,
+                                                     on_users_monitor_changed);
 
-        if (daemon->priv->passwd_monitor != NULL) {
-                g_signal_connect (daemon->priv->passwd_monitor,
-                                  "changed",
-                                  G_CALLBACK (on_passwd_monitor_changed),
-                                  daemon);
-        } else {
-                g_warning ("Unable to monitor %s: %s", PATH_PASSWD, error->message);
-                g_error_free (error);
-        }
-        if (daemon->priv->shadow_monitor != NULL) {
-                g_signal_connect (daemon->priv->shadow_monitor,
-                                  "changed",
-                                  G_CALLBACK (on_passwd_monitor_changed),
-                                  daemon);
-        } else {
-                g_warning ("Unable to monitor %s: %s", PATH_SHADOW, error->message);
-                g_error_free (error);
-       }
-        if (daemon->priv->gdm_monitor != NULL) {
-                g_signal_connect (daemon->priv->gdm_monitor,
-                                  "changed",
-                                  G_CALLBACK (on_gdm_monitor_changed),
-                                  daemon);
-        } else {
-                g_warning ("Unable to monitor %s: %s", PATH_GDM_CUSTOM, error->message);
-                g_error_free (error);
-        }
+        daemon->priv->wtmp_monitor = setup_monitor (daemon,
+                                                    wtmp_helper_get_path_for_monitor (),
+                                                    on_users_monitor_changed);
+
+        daemon->priv->gdm_monitor = setup_monitor (daemon,
+                                                   PATH_GDM_CUSTOM,
+                                                   on_gdm_monitor_changed);
 
         queue_reload_users (daemon);
         queue_reload_autologin (daemon);
@@ -798,6 +559,8 @@ daemon_finalize (GObject *object)
 
         g_hash_table_destroy (daemon->priv->users);
 
+        g_hash_table_unref (daemon->priv->extension_ifaces);
+
         G_OBJECT_CLASS (daemon_parent_class)->finalize (object);
 }
 
@@ -927,7 +690,7 @@ daemon_local_find_user_by_name (Daemon      *daemon,
 
         pwent = getpwnam (name);
         if (pwent == NULL) {
-                g_debug ("unable to lookup name %s", name);
+                g_debug ("unable to lookup name %s: %s", name, g_strerror (errno));
                 return NULL;
         }
 
@@ -939,6 +702,12 @@ daemon_local_find_user_by_name (Daemon      *daemon,
         return user;
 }
 
+User *
+daemon_local_get_automatic_login_user (Daemon *daemon)
+{
+        return daemon->priv->autologin;
+}
+
 static gboolean
 daemon_find_user_by_id (AccountsAccounts      *accounts,
                         GDBusMethodInvocation *context,
@@ -1024,17 +793,12 @@ finish_list_cached_users (gpointer user_data)
                 uid = user_get_uid (user);
                 shell = user_get_shell (user);
 
-                if (user_get_system_account (user)) {
-                        g_debug ("user %s %ld is system account, so excluded\n", name, (long) uid);
+                if (!user_classify_is_human (uid, name, shell, NULL)) {
+                        g_debug ("user %s %ld excluded", name, (long) uid);
                         continue;
                 }
 
-                if (daemon_local_user_is_excluded (data->daemon, name, shell)) {
-                        g_debug ("user %s %ld excluded\n", name, (long) uid);
-                        continue;
-                }
-
-                g_debug ("user %s %ld not excluded\n", name, (long) uid);
+                g_debug ("user %s %ld not excluded", name, (long) uid);
                 g_ptr_array_add (object_paths, (gpointer) user_get_object_path (user));
         }
         g_ptr_array_add (object_paths, NULL);
@@ -1074,6 +838,24 @@ daemon_get_daemon_version (AccountsAccounts *object)
     return VERSION;
 }
 
+static void
+cache_user (Daemon *daemon,
+            User   *user)
+{
+        gchar       *filename;
+        const char  *user_name;
+
+        /* Always use the canonical user name looked up */
+        user_name = user_get_user_name (user);
+
+        filename = g_build_filename (USERDIR, user_name, NULL);
+        if (!g_file_test (filename, G_FILE_TEST_EXISTS)) {
+                user_save (user);
+        }
+
+        g_free (filename);
+}
+
 typedef struct {
         gchar *user_name;
         gchar *real_name;
@@ -1116,7 +898,7 @@ daemon_create_user_authorized_cb (Daemon                *daemon,
         argv[3] = cd->real_name;
         if (cd->account_type == ACCOUNT_TYPE_ADMINISTRATOR) {
                 argv[4] = "-G";
-                argv[5] = "wheel";
+                argv[5] = ADMIN_GROUP;
                 argv[6] = "--";
                 argv[7] = cd->user_name;
                 argv[8] = NULL;
@@ -1140,6 +922,9 @@ daemon_create_user_authorized_cb (Daemon                *daemon,
 
         user = daemon_local_find_user_by_name (daemon, cd->user_name);
         user_update_local_account_property (user, TRUE);
+        user_update_system_account_property (user, FALSE);
+
+        cache_user (daemon, user);
 
         accounts_accounts_complete_create_user (NULL, context, user_get_object_path (user));
 }
@@ -1178,9 +963,6 @@ daemon_cache_user_authorized_cb (Daemon                *daemon,
                                  gpointer               data)
 {
         const gchar *user_name = data;
-        GError      *error = NULL;
-        gchar       *filename;
-        gchar       *comment;
         User        *user;
 
         sys_log (context, "cache user '%s'", user_name);
@@ -1192,23 +974,9 @@ daemon_cache_user_authorized_cb (Daemon                *daemon,
                 return;
         }
 
-        /* Always use the canonical user name looked up */
-        user_name = user_get_user_name (user);
+        user_update_system_account_property (user, FALSE);
 
-        filename = g_build_filename (USERDIR, user_name, NULL);
-        if (!g_file_test (filename, G_FILE_TEST_EXISTS)) {
-                comment = g_strdup_printf ("# Cached file for %s\n\n", user_name);
-                g_file_set_contents (filename, comment, -1, &error);
-                g_free (comment);
-
-                if (error != NULL) {
-                        g_warning ("Couldn't write user cache file: %s: %s",
-                                   filename, error->message);
-                        g_error_free (error);
-                }
-        }
-
-        g_free (filename);
+        cache_user (daemon, user);
 
         accounts_accounts_complete_cache_user (NULL, context, user_get_object_path (user));
 }
@@ -1295,7 +1063,7 @@ daemon_uncache_user (AccountsAccounts      *accounts,
 }
 
 typedef struct {
-        gint64 uid;
+        uid_t uid;
         gboolean remove_files;
 } DeleteUserData;
 
@@ -1322,6 +1090,19 @@ daemon_delete_user_authorized_cb (Daemon                *daemon,
 
         sys_log (context, "delete user '%s' (%d)", pwent->pw_name, ud->uid);
 
+        if (daemon->priv->autologin != NULL) {
+                User *user;
+
+                user = daemon_local_find_user_by_id (daemon, ud->uid);
+
+                g_assert (user != NULL);
+
+                if (daemon->priv->autologin == user) {
+                        daemon_local_set_automatic_login (daemon, user, FALSE, NULL);
+                }
+
+        }
+
         filename = g_build_filename (USERDIR, pwent->pw_name, NULL);
         g_remove (filename);
         g_free (filename);
@@ -1339,9 +1120,10 @@ daemon_delete_user_authorized_cb (Daemon                *daemon,
                 argv[5] = NULL;
         }
         else {
-                argv[1] = "--";
-                argv[2] = pwent->pw_name;
-                argv[3] = NULL;
+                argv[1] = "-f";
+                argv[2] = "--";
+                argv[3] = pwent->pw_name;
+                argv[4] = NULL;
         }
 
         error = NULL;
@@ -1364,13 +1146,13 @@ daemon_delete_user (AccountsAccounts      *accounts,
         Daemon *daemon = (Daemon*)accounts;
         DeleteUserData *data;
 
-        if (uid == 0) {
+        if ((uid_t)uid == 0) {
                 throw_error (context, ERROR_FAILED, "Refuse to delete root user");
                 return TRUE;
         }
 
         data = g_new0 (DeleteUserData, 1);
-        data->uid = uid;
+        data->uid = (uid_t)uid;
         data->remove_files = remove_files;
 
         daemon_local_check_auth (daemon,
@@ -1603,6 +1385,12 @@ daemon_local_set_automatic_login (Daemon    *daemon,
         return TRUE;
 }
 
+GHashTable *
+daemon_get_extension_ifaces (Daemon *daemon)
+{
+  return daemon->priv->extension_ifaces;
+}
+
 static void
 get_property (GObject    *object,
               guint       prop_id,
